import atexit
import os
import sys
import uuid
from types import FunctionType
from typing import Union, Tuple, List

from .globs import _app_ctx_stack
from .util import ContextStore, utils
from . import __meta__
from .http import HttpRequest, NotFound, ServerError
from .middleware import MiddlewareManager
from .routes import Collector


class AppContext:
    def __init__(self, app):
        self.app = app
        self.store = ContextStore(_app_ctx_stack)
        self.ref_count = 0

    def push(self):
        top = _app_ctx_stack.top
        if not top or top != self:
            _app_ctx_stack.push(self)
        self.ref_count += 1

    def pop(self):
        if _app_ctx_stack.top != self:
            return
        self.ref_count -= 1
        if self.ref_count > 0:
            return
        _app_ctx_stack.pop()
        del self.store

    def __enter__(self):
        self.push()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop()


def print_meta():
    print(__meta__.text_img)


class App:
    # APP 实例集合
    INSTANCES = {}

    def __init__(self,
                 app_root: str,
                 app_id: str = None,
                 api_prefix='api',
                 debug=True,
                 append_slash=False,
                 strict_mode=False,
                 api_page_enabled=None,
                 api_page_name: str = None,
                 api_page_expanded=True,
                 api_page_cache=None,
                 api_page_addition=None,
                 api_page_header=None,
                 api_page_footer=None,
                 api_page_assets: tuple = None,
                 allowed_route_meta: dict = None,
                 favicon: str = None
                 ):
        """

        :param app_root: 应用的根目录
        :param app_id: 全局的唯一 id, 用于标识一个APP。
        :param api_prefix: API接口URL前缀
        :param debug: 是否启用调试模式
        :param append_slash: 是否在URL末尾使用 / 符号
        :param strict_mode: 是否启用严格模式。在严格模式下，不允许在请求时携带未声明的参数
        :param api_page_enabled: 只否启用接口页面
        :param api_page_name: 接口页面名称
        :param api_page_expanded: 是否在页面加载后自动展开接口列表
        :param api_page_cache: 是否缓存接口数据
        :param api_page_addition: 接口页面上要展示的接口的附加信息函数，其接收一个 dict 类型的参数 route_info
        :param api_page_header: 接口页面上，需要自定义展示在页面顶部的信息
        :param api_page_footer: 接口页面上，需要自定义展示在页面底部的信息
        :param api_page_assets: 接口页面上，需要自定义的JS和CSS以及其它资源 (通过 static 托管，此接口仅填写地址)
        :param allowed_route_meta: 路由装饰器上，指定允许使用的自定义元数据参数。当此值为 None 时，不限制。
        """
        print_meta()

        # 更新系统路径
        if app_root not in sys.path:
            sys.path.insert(0, app_root)

        from .config import AppConfig
        from .routes import ApiPage
        from .routes import Router
        from .util import Logger

        # 上下文需要先创建，以便后续的模块能直接使用
        self.context = AppContext(self)
        self.context.push()

        self.id = app_id or str(uuid.uuid4())
        self._logger = Logger(app_id)

        self.INSTANCES[self.id] = self

        self.config = AppConfig(self.id, app_root, debug, api_prefix, append_slash,
                                strict_mode, api_page_enabled, api_page_name,
                                api_page_expanded, api_page_cache, api_page_addition,
                                api_page_header, api_page_footer, api_page_assets,
                                allowed_route_meta)
        self.config.middleware_manager = MiddlewareManager(self.config)
        self.api_prefix = api_prefix
        self.router = Router(self.config)
        self.api_page = ApiPage(self.config)

        self.custom_url_map = {}

        self.favicon = favicon

        from werkzeug.routing import Map, Rule
        self.url_map = Map([
            Rule('/favicon.ico', endpoint='favicon'),
            Rule('/%s%s' % (api_prefix, '/' if append_slash else ''), endpoint='_api_page'),
            Rule('/%s/api.json' % api_prefix, endpoint='_api_page'),
            Rule('/%s/export' % api_prefix, endpoint='_api_page'),
            Rule('/%s/<path:entry>%s' % (api_prefix, '/' if append_slash else ''), endpoint='entry_only')
        ])

        Collector.create(app_id, app_root, append_slash)

    def dispose(self):
        self._logger.info('App "%s" is disposing' % self.id)

        if self.id in self.INSTANCES:
            self.INSTANCES.pop(self.id)

        self.config.middleware_manager.handle_shutdown()

        self.context.pop()
        self.config.dispose()
        Collector.destroy(self.id)

        self._logger.info('App "%s" exited' % self.id)

    def handle_wsgi_request(self, environ, start_response):
        """
        接收并处理来自 wsgi 的请求
        :param environ:
        :param start_response:
        :return:
        """

        request = None
        response = None
        try:
            request = HttpRequest(environ, self)
            request.context().push()

            response = self.config.middleware_manager.handle_coming(request)
            if not response:
                adapter = self.url_map.bind_to_environ(environ)

                endpoint, values = adapter.match()
                if endpoint == '_api_page':
                    response = self.api_page.dispatch(request)
                elif endpoint == 'entry_only':
                    response = self.router.dispatch(request, values['entry'])
                elif endpoint in self.custom_url_map:
                    response = self.custom_url_map[endpoint](request, **values)
                elif endpoint == 'favicon':
                    if self.favicon is None:
                        response = NotFound()
                    else:
                        from .http import FileResponse
                        response = FileResponse(os.path.join(self.config.ROOT, self.favicon))
                else:
                    response = NotFound()
        except Exception as e:
            # 忽略静态资源请求的 .js.map/.css.map 文件
            if request and (request.path.lower().endswith('.js.map') or request.path.lower().endswith('.css.map')):
                response = NotFound()
            else:
                msg = utils.get_exception_info(e)
                if request:
                    msg = 'Error occurred during handling request %r:\n\t%s' % (request.path, msg)

                self._logger.error(msg)

                from werkzeug.exceptions import NotFound as SuperNotFound
                if isinstance(e, SuperNotFound):
                    response = NotFound()
                elif self.config.debug:
                    response = ServerError(msg.replace('<', '&lt;').replace('>', '&gt;'))
                else:
                    response = ServerError()
        finally:
            result = self.config.middleware_manager.handle_leaving(request, response)
            if result:
                response = result
            request.context().pop()
            import werkzeug
            sv = sys.version_info
            response.headers['server'] = '%s/%s werkzeug/%s Python/%s' % (
                __meta__.name, __meta__.version, werkzeug.__version__, '%s.%s.%s' % (sv[0], sv[1], sv[2])
            )
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        """
        wsgi 入口
        :param environ:
        :param start_response:
        :return:
        """
        if not self.config.static_map:
            return self.handle_wsgi_request(environ, start_response)

        from werkzeug.middleware.shared_data import SharedDataMiddleware
        return SharedDataMiddleware(self.handle_wsgi_request, self.config.static_map)(
            environ, start_response)

    def startup(self, host=None, port=9127, threaded=True, **kwargs):
        """
        开发时使用的服务器，一般来说，应仅用于开发
        :param host:
        :param port:
        :param threaded:
        :param kwargs: 适用于 werkzeug 的 run_simple 函数的其它参数
        :return:
        """
        from .util import helper
        from werkzeug.serving import run_simple

        debug = self.config.debug

        if debug:
            helper.print_meta('dev-server *DEBUG MODE*')
        else:
            helper.print_meta('dev-server')

        # 检查启动参数
        cmd_host = None
        cmd_port = None
        argv = sys.argv
        if len(argv) > 1:
            server_arg = argv[1].split(':')
            if len(server_arg) == 1:
                self._logger.warning('Invalid startup argument "%s" (ignored):' % argv[1])
            else:
                cmd_host = server_arg[0]
                cmd_port = int(server_arg[1])

        # 检查环境变量
        env_host = os.environ.get('RESTFX_HOST')
        env_port = os.environ.get('RESTFX_PORT')

        if env_host is not None:
            host = env_host
            self._logger.info('Using host from environment variable RESTFX_HOST: %s' % host)
        elif cmd_host:
            host = cmd_host
            self._logger.info('Using host from command line host: %s' % host)
        else:
            self._logger.info('Using host in code: %s' % host)
        if env_port is not None:
            port = int(env_port)
            self._logger.info('Using port from environment variable RESTFX_PORT: %s' % port)
        elif cmd_host:
            port = cmd_port
            self._logger.info('Using port from command line port: %s' % port)
        else:
            self._logger.info('Using port in code: %s' % port)

        # 检查端口是否被占用
        # if utils.is_port_used(port):
        #     self._logger.warning('The port "%s" is not available, pick another one instead.' % port)
        #     import sys
        #     sys.exit(1)

        # 支持 空串和 * 标志
        if host in [None, '', '*']:
            host = '0.0.0.0'

        if self.config.api_page_enabled:
            if host == '0.0.0.0':
                ips = utils.get_ip_list()
            else:
                ips = [host]

            protocol = 'https' if kwargs.get('ssl_context') else 'http'

            print(' * Table of APIs:')
            for ip in ips:
                print('\t- %s://%s:%s/%s%s' % (
                    protocol, ip, port, self.api_prefix, '/' if self.config.append_slash else ''
                ))

        run_simple(host, port, self, use_debugger=False, use_reloader=debug, threaded=threaded, **kwargs)

    def register_types(self, *types):
        """
        注册路由参数上使用的类型，目前仅支持枚举类型
        :param types:
        :return:
        """
        collector = Collector.get(self.id)
        for type_item in types:
            collector.global_types[type_item.__name__] = type_item

        return self

    def collect(self):
        """
        收集路由
        :return:
        """
        return Collector.get(self.id).collect(self.config.routes_map)

    def persist(self, filename: str = 'dist/routes_map.py', encoding='utf8'):
        """
        持久化路由信息到文件
        :param encoding: 文件编码
        :param filename: 输出的路由文件名称
        :return:
        """
        routes_file = os.path.basename(filename)
        routes_dir = os.path.dirname(filename)
        return Collector.get(self.id).persist(self.config.routes_map, routes_file, routes_dir, encoding)

    def set_logger(self, logger):
        """
        设置自定义的 logger
        :param logger:
        :return:
        """
        self._logger.custom_logger = logger
        return self

    def map_urls(self, urls_map: dict):
        """
        自定义的 url 映射，这些映射不会经过路由处理，而是直接来自 werkzeug
        :param urls_map:
        :return:
        """
        from werkzeug.routing import Rule

        for url in urls_map:
            self.custom_url_map[url] = urls_map[url]
            self.url_map.add(Rule(url, endpoint=url))
        return self

    def map_routes(self, routes_map: dict):
        """
        注册路由映射表
        :param routes_map:
        :return:
        """
        for path in routes_map:
            if not path:
                raise Exception('Empty route map prefix value')
            pkg = routes_map[path]
            pkg_path = os.path.join(self.config.ROOT, pkg.replace('.', '/'))
            if not os.path.exists(pkg_path):
                self._logger.warning('Route map "%s" does not exist, path=%s' % (pkg, pkg_path))
            self.config.routes_map[path] = pkg
        return self

    def map_static(self, static_map: dict):
        """
        设置静态资源目录映射
        :param static_map:
        :return:
        """
        for prefix in static_map:
            # 静态资源的前缀，必须以 / 开头
            if prefix[0] != '/':
                self._logger.warning(
                    'Invalid prefix of static map, it should start with "/": %s' % prefix)
                continue
            if prefix.lower() == '/internal_assets':
                self._logger.warning(
                    'The prefix of static map is reserved, pick another one please: %s' % prefix)
                continue
            target_path = static_map[prefix]
            abs_target_path = os.path.abspath(os.path.join(self.config.ROOT, target_path))
            # 无论目录是否存在都注册，因为静态目录可能是动态生成的
            self.config.static_map[prefix] = abs_target_path
            if os.path.exists(abs_target_path):
                self._logger.debug('Map static url %s to path %s' % (prefix, abs_target_path))
            else:
                self._logger.warning(
                    'The target path of static url %s not found: %s' % (prefix, abs_target_path))
        return self

    def register_routes(self, routes: list):
        """
        注册路由列表。一般来说，应当通过以下方法调用：

        from dist import routes_map

        app.register_routes(routes_map.routes)

        :param routes: 其每一项都应该是一个 list, 元素依次为 method: str, path: str, handler: FunctionType
        :return:
        """
        for handler, api_info in routes:
            from .util.func_util import FunctionDescription

            method = api_info['method']
            path = api_info['path']

            rid = '%s#%s' % (path, method.lower())
            if rid in self.router.production_routes:
                self._logger.warning('Duplicated route %s %s' % (method, path))

            desc = FunctionDescription(handler)
            self.router.production_routes[rid] = {
                'func': handler,
                'args': desc.arguments
            }
            self.config.docs.append(api_info)

        return self

    def register_middleware(self, *middlewares, index: int = -1):
        """
        注册中间件，注册的中间件将按顺序执行
        :param middlewares: 中间件实例列表
        :param index: 要将包插入的位置，指定为 -1 表示放到最后
        :return:
        """
        from .middleware import MiddlewareBase

        for middleware in middlewares:
            assert isinstance(middleware, MiddlewareBase)
            if index == -1:
                self.config.middlewares.append(middleware)
                self.config.reversed_middlewares.insert(0, middleware)
            else:
                self.config.middlewares.insert(index, middleware)
                if index == 0:
                    self.config.reversed_middlewares.append(middleware)
                else:
                    self.config.reversed_middlewares.insert(-index, middleware)
                # 每添加一个，将 index 后移动一个位置
                index += 1
            # 调用中间件，标记中间件的启动
            middleware.on_startup(self)

        return self

    def inject(self, **kwargs):
        """
        全局注入接口
        :param kwargs:
        :return:
        """
        self.config.injections.update(**kwargs)
        return self

    def register_plugins(self, *plugins):
        """
        注册插件，插件没有顺序之分
        :param plugins: 插件实例列表
        :return:
        """
        from .plugin import PluginBase

        for plugin in plugins:
            assert isinstance(plugin, PluginBase)
            self.config.plugins.append(plugin)
            # 调用插件，标记中间件的启动
            plugin.setup(self)

        return self

    def register_enums(self, enum_types: Union[list, tuple], ignore_non_enum=True):
        """
        注册枚举类型
        :param enum_types: 要注册的枚举类型
        :param ignore_non_enum: 是否忽略慧枚举类型
        :return:
        """
        from enum import Enum, IntEnum, Flag, IntFlag

        for enum_type in enum_types:
            if not ignore_non_enum:
                assert issubclass(enum_type, Enum)
            if enum_type in (Enum, IntEnum, Flag, IntFlag):
                # 忽略系统的枚举类型
                # self._logger.debug('Ignore non-user defined of enum type: ' + enum_type.__name__)
                continue
            if sys.version_info >= (3, 11):
                from enum import StrEnum, FlagBoundary
                if enum_type in (StrEnum, FlagBoundary):
                    continue
            if enum_type in self.config.enum_types:
                continue
            self.config.enum_types.append(enum_type)

        return self

    def scan_routes(self, prefix: str = 'app_', sub_dir='entry', ignores: Union[str, Tuple[str], List[str]] = None):
        """
        自动扫描路由目录
        :param prefix: 要匹配的目录前缀，除去前缀后剩下的部分，会被作为 url 上的路径名称，如： app_demo，对应的 url 为 /api/demo
        :param sub_dir: 匹配的目录下的子目录名称，通常，API文件会放到独立目录下；不指定时扫描整个目录
        :param ignores: 要忽略的目录前缀(或者完整)名称列表
        :return:
        """
        prefix = prefix.lower()

        if isinstance(ignores, str):
            ignores = (ignores,)

        def is_ignore(matching_dir_name):
            """
            检查项是否需要忽略
            :param matching_dir_name:
            :return:
            """
            if not ignores:
                return False

            for ignore_prefix in ignores:
                if matching_dir_name.startswith(ignore_prefix.lower()):
                    return True

            return False

        # 扫描目录，自动加载应用的模块 (app_开头)
        routes_map = {}
        for dir_name in os.listdir(self.config.ROOT):
            lower_dir_name = dir_name.lower()

            if is_ignore(lower_dir_name):
                self._logger.info('Ignore directory: ' + dir_name)
                continue
            if not dir_name.startswith(prefix):
                self._logger.info('Mismatch directory: ' + dir_name)
                continue

            self._logger.info('Match directory: ' + dir_name)

            item_path = os.path.join(self.config.ROOT, dir_name)

            if not os.path.isfile(os.path.join(item_path, '__init__.py')):
                self._logger.warning('Invalid directory: "%s" is not a package' % dir_name)
                continue

            if sub_dir:
                entry_path = os.path.join(item_path, sub_dir)
                if not os.path.isfile(os.path.join(entry_path, '__init__.py')):
                    self._logger.warning('Invalid directory: "%s%s%s" is not a package' % (
                        dir_name, os.path.sep, sub_dir
                    ))
                    continue

            url_path_name = lower_dir_name[len(prefix):]
            routes_map[url_path_name] = '%s.%s' % (dir_name, sub_dir) if sub_dir else dir_name

        if not routes_map:
            raise Exception('Cannot find any directory which\'s name startswith "%s"' % prefix)

        self.map_routes(routes_map)

    @classmethod
    def register_command(cls, command: str, handler: FunctionType, description: str):
        from .commands import commands
        commands.register(command, handler, description)

    def test_command(self):
        from .commands import execute
        return execute(*sys.argv, working_dir=self.config.ROOT)

    @classmethod
    def get(cls, app_id: str):
        """
        获取指定的应用
        :param app_id:
        :return:
        :rtype: App
        """
        return cls.INSTANCES.get(app_id)

    @staticmethod
    def current():
        """
        获取当前的应用
        :return:
        """
        from . import globs
        return globs.current_app

    @staticmethod
    def store():
        """
        获取当前应用的存储对象
        :return:
        """
        from . import globs
        return globs.app_store


@atexit.register
def _destroy_apps():
    # Parse the ids into list to avoid the error "RuntimeError: dictionary changed size during iteration"
    app_ids = list(App.INSTANCES.keys())
    for app_id in app_ids:
        app = App.get(app_id)
        app.dispose()



from ...globs import current_app
from ...http import HttpResponse, NotFound
from ...middleware import MiddlewareBase
from ...routes.route_resolver import RouteResolver


class OptionsMiddleware(MiddlewareBase):
    """
    提供 options method 请求支持
    """

    # 可用的 method，即支持的 method 列表
    AVAILABLE_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'PATCH']

    def __init__(self,
                 allow_origin=None,
                 allow_methods=None,
                 allow_headers=None,
                 allow_credentials=None,
                 max_age=None,
                 content_security_policy=None,
                 content_security_policy_report_only=None,
                 access_control_expose_headers=None,
                 cross_origin_opener_policy=None,
                 cross_origin_embedder_policy=None,
                 ):
        self.allow_origin = allow_origin
        self.allow_methods = allow_methods
        self.allow_headers = allow_headers
        self.allow_credentials = allow_credentials
        self.max_age = max_age
        self.content_security_policy = content_security_policy
        self.content_security_policy_report_only = content_security_policy_report_only
        self.access_control_expose_headers = access_control_expose_headers
        self.cross_origin_opener_policy = cross_origin_opener_policy
        self.cross_origin_embedder_policy = cross_origin_embedder_policy

    def on_coming(self, request):
        if request.method != 'OPTIONS':
            return

        allow_methods = self.allow_methods or self.get_allow_methods(request)
        if not allow_methods:
            return NotFound()

        response = HttpResponse()
        response.access_control_allow_methods = allow_methods

        if self.allow_origin is not None:
            response.access_control_allow_origin = self.allow_origin

        if self.allow_credentials is not None:
            response.access_control_allow_credentials = self.allow_credentials

        if self.allow_headers is not None:
            response.access_control_allow_headers = self.allow_headers

        if self.max_age is not None:
            response.access_control_max_age = self.max_age

        if self.content_security_policy is not None:
            response.content_security_policy = self.content_security_policy

        if self.content_security_policy_report_only is not None:
            response.content_security_policy_report_only = self.content_security_policy_report_only

        if self.access_control_expose_headers is not None:
            response.access_control_expose_headers = self.access_control_expose_headers

        if self.cross_origin_opener_policy is not None:
            response.cross_origin_opener_policy = self.cross_origin_opener_policy

        if self.cross_origin_embedder_policy is not None:
            response.cross_origin_embedder_policy = self.cross_origin_embedder_policy

        return response

    @classmethod
    def get_allow_methods(cls, request):
        app = current_app

        path = request.path
        entry = path[len('/' + app.api_prefix):]

        allow_methods = []

        if app.config.debug:
            # 调试模式时，从源码查找可用的路由
            entry = entry.lstrip('/')
            resolver = RouteResolver(
                app.config,
                app.router.entry_cache,
                request.method, entry
            )

            # 分别尝试不同 method 的路由是否存在
            for method in cls.AVAILABLE_METHODS:
                # 定义时，路由必须使用小写，所以处理成小写形式再去匹配
                if cls.has_method(resolver, method.lower()):
                    allow_methods.append(method)
        else:
            # 分别尝试不同 method 的路由是否存在
            for method in cls.AVAILABLE_METHODS:
                # 定义时，路由必须使用小写，所以处理成小写形式再去匹配
                rid = '%s#%s' % (entry, method.lower())
                if rid in app.router.production_routes:
                    allow_methods.append(method)

        return allow_methods

    @classmethod
    def has_method(cls, resolver, method: str):
        result = resolver.resolve(method)
        if not result:
            return False

        if isinstance(result, HttpResponse):
            return False

        return True



import os

from werkzeug.local import LocalStack

from . import __meta__
from .util import Logger


class AppConfig:
    """
    应用的配置
    """

    # 所有应用的配置集合
    _CONFIGS = {}

    def __init__(self, app_id: str,
                 app_root: str,
                 debug: bool,
                 api_prefix: str,
                 append_slash: bool,
                 strict_mode: bool,
                 api_page_enabled: bool,
                 api_page_name: str,
                 api_page_expanded: bool,
                 api_page_cache: bool,
                 api_page_addition,
                 api_page_header: str,
                 api_page_footer: str,
                 api_page_assets: tuple,
                 allowed_route_meta: dict,
                 ):
        """

        """
        self._CONFIGS[app_id] = self
        self.app_id = app_id
        self.api_prefix = api_prefix
        self.local = LocalStack()
        # 是否启用DEBUG模式
        self.debug = debug
        # 工作目录
        self.ROOT = app_root
        self.append_slash = append_slash
        self.strict_mode = strict_mode
        self.middleware_manager = None
        self.middlewares = []
        """
        注册的中间件实例集合
        :type: List[MiddlewareBase]
        """
        self.reversed_middlewares = []
        """
        :type: List[MiddlewareBase]
        """
        self.plugins = []
        """
        注册的插件实例集合
        :type: List[PluginBase]
        """

        self.routes = []
        """
        注册的路由列表
        """

        self.docs = []
        """
        注册的路由列表对应的文档信息
        """

        self.enum_types = []
        """
        系统使用到的枚举类型集合
        """

        # 路由映射表，其键为请求的路径，其值为映射的目录
        self.routes_map = {}

        self.static_map = {}

        # 注入数据/函数集合
        # 其中存放将被注入到路由函数参数列表上的数据/函数
        self.injections = {}

        # 是否启用 API 页面
        self.api_page_enabled = debug if api_page_enabled is None else api_page_enabled
        self.api_page_name = api_page_name or 'An awesome %s project' % __meta__.name
        self.api_page_expanded = api_page_expanded
        # 是否缓存API页面的 html 文件 和 接口数据
        self.api_page_cache = not debug if api_page_cache is None else api_page_cache
        self.api_page_header = api_page_header
        self.api_page_footer = api_page_footer
        self.api_page_addition = api_page_addition
        self.api_page_assets = api_page_assets
        self.allowed_route_meta = allowed_route_meta

        if self.api_page_enabled:
            self.static_map['/internal_assets'] = os.path.join(os.path.dirname(__file__), 'internal_assets')

    def dispose(self):
        if self.app_id in self._CONFIGS:
            self._CONFIGS.pop(self.app_id)

        logger = Logger._LOGGERS[self.app_id]

        for plugin in self.plugins:
            try:
                plugin.destroy()
            except Exception as e:
                logger.error('Failed to destroy plugin %r' % plugin.__name__, e)

        for middleware in self.reversed_middlewares:
            # logger.debug('Disposing middleware %r' % type(middleware).__name__)
            try:
                middleware.dispose()
            except Exception as e:
                Logger.current().error('Failed to dispose middleware %r' % middleware.__name__, e)

        del self.middlewares
        del self.reversed_middlewares
        del self.plugins
        del self.routes
        del self.enum_types
        del self.api_page_assets
        del self.routes_map
        del self.static_map

    @classmethod
    def get(cls, app_id: str):
        """

        :param app_id:
        :return:
        :rtype: AppConfig
        """
        return cls._CONFIGS.get(app_id)

    @classmethod
    def current(cls):
        from . import globs
        app_id = globs.current_app.id
        return cls.get(app_id)



import time
from abc import ABC, abstractmethod
from threading import Timer
from types import FunctionType
from typing import List, Optional

from .session import HttpSession


class ISessionProvider(ABC):
    """
    session 提供者基类，自定义 session 提供者时应该继承此类
    此类不可被实例化
    """

    def __init__(self, expired: int, check_interval: int, auto_clear: bool, on_expired: FunctionType = None):
        """

        :param expired: 过期时长，单位为秒，默认为 10分钟。指定为 0表示不会自动过期
        :param check_interval: 检查 session 是否过期的周期，单位为秒，指定为 0 时不检查
        :param auto_clear: 是否在停止/启动时，自动清空 session 存储
        :param on_expired: 在 session 过期时，触发
        """
        from ..util import Logger
        self.logger = Logger.current()
        self.auto_clear = auto_clear

        self.expired = 10 * 60 if expired is None else expired
        self.on_expired = on_expired
        self.check_interval = check_interval
        self.timer = self.run_timer()

    def run_timer(self):
        # 不会过期，不用启动定时器
        if self.expired <= 0 or self.check_interval <= 0:
            return None

        timer = Timer(self.check_interval, self.drop_expired_session)
        timer.start()
        return timer

    def drop_expired_session(self):
        time_before = time.time() - self.expired
        expired_sessions = self.get_expired_session(time_before)
        for session_id in expired_sessions:
            if self.on_expired:
                self.on_expired(self.get(session_id))
            # self.logger.debug('Drop expired session: ' + session_id)
            self.remove(session_id)

        self.timer = self.run_timer()

    def is_expired(self, session: HttpSession) -> bool:
        if self.expired > 0:
            return time.time() - session.last_access_time > self.expired
        return False

    def create(self, session_id: str) -> HttpSession:
        session = HttpSession(session_id, self.set, self.remove)
        # 实际存储了数据再创建
        # self.set(session)
        return session

    @abstractmethod
    def get_expired_session(self, time_before: float) -> List[str]:
        """
        查询已经过期的 session 的 id
        :param time_before: 在此时间之前的 session 即为过期
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def get(self, session_id: str) -> Optional[HttpSession]:
        """
        获取指定 id 的 session
        :param session_id:
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def set(self, session: HttpSession):
        """
        添加或更新指定的 session
        :param session:
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def exists(self, session_id: str) -> bool:
        """
        检索指定 id 的 session 是否存在
        :param session_id:
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def remove(self, session_id: str):
        """
        销毁一个 session
        :param session_id:
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def clear(self):
        """
        清空所有 session
        :return:
        """
        raise NotImplemented()

    def dispose(self):
        """
        回收 session provider
        :return:
        """
        if self.timer and self.timer.is_alive():
            self.timer.cancel()
            self.timer.join()
        if self.auto_clear:
            self.clear()

    def __del__(self):
        self.dispose()


class IDbSessionProvider(ISessionProvider):
    def __init__(self, db_options: dict, expired: int, check_interval: int, auto_clear=False,
                 on_expired: FunctionType = None):
        super().__init__(expired, check_interval, auto_clear, on_expired)

        self.db_options = db_options
        self.is_db_available = False

    def drop_expired_session(self):
        if self.is_db_available:
            super(IDbSessionProvider, self).drop_expired_session()

    def parse(self, data: dict) -> HttpSession:
        """
        用于将字典转换成 HttpSession 对象
        :param data:
        :return:
        """
        import pickle
        session = HttpSession(data['id'], self.set, self.remove)
        session.creation_time = int(data['creation_time'])
        session.last_access_time = int(data['last_access_time'])
        session.store = pickle.loads(data['store'])

        return session

    def set(self, session: HttpSession):
        import pickle
        self.upsert(session.id,
                    session.creation_time,
                    session.last_access_time,
                    pickle.dumps(session.store))

    @abstractmethod
    def upsert(self, session_id: str, creation_time: float, last_access_time: float, store: bytes):
        raise NotImplemented()



import os
import sys

from .commands import (
    run_command,
    register,
    command_help,
    command_create,
    command_version, command_genid
)

register('help', command_help, 'Show this help message')
register('version', command_version, 'Show the version information')
register('create', command_create, 'Create restfx project structure', 'project-name')
register('genid', command_genid, 'Generate a new id for app')


def execute(*args, **kwargs):
    args = list((args or sys.argv)[1:])

    if 'working_dir' in kwargs:
        working_dir = kwargs['working_dir']
    else:
        working_dir = os.path.abspath(os.path.curdir)

    if len(args) == 0:
        return False
        # return run_command(working_dir, 'help')

    command_name = args.pop(0)
    return run_command(working_dir, command_name, *args)


__all__ = [
    'run_command',
    'register',
    'command_help',
    'command_create',
    'command_version',
    'command_genid',
    'execute'
]



import time

from .. import MiddlewareBase


class TimetickMiddleware(MiddlewareBase):
    """
    用于记录在路由的不同阶段的耗时情况
    """

    KEY = '__timetick__'
    HEADER = 'restfx-timetick'

    def __init__(self):
        pass

    def _get_timetick(self):
        return time.perf_counter() * 1000

    def force_run_method(self, method: int):
        return True

    def on_coming(self, request):
        request.set(self.KEY, [self._get_timetick()])

    def process_request(self, request, meta):
        request.get(self.KEY).append(self._get_timetick())

    def process_invoke(self, request, meta, args):
        request.get(self.KEY).append(self._get_timetick())

    def process_return(self, request, meta, data):
        request.get(self.KEY).append(self._get_timetick())

    def process_response(self, request, meta, response):
        request.get(self.KEY).append(self._get_timetick())

    def on_leaving(self, request, response):
        timetick = request.get(self.KEY)
        timetick.append(self._get_timetick())

        total = round(timetick[-1] - timetick[0], 3)
        detail = []
        i = 1
        prev = None
        for tt in timetick:
            if prev is not None:
                detail.append('%s/%sms' % (i, round(tt - prev, 3)))
                i += 1

            prev = tt

        response.headers[self.HEADER] = '%sms; %s' % (round(total, 3), ', '.join(detail))



from abc import ABC, abstractmethod


class PluginBase(ABC):
    """
    插件基类
    """

    def __init__(self, name: str = None, version: str = None, description: str = None):
        """

        :param name: 指定插件名称
        :param version: 指定插件版本
        :param description: 指定插件描述
        """
        self.name = name
        self.version = version
        self.description = description

    @abstractmethod
    def setup(self, app):
        """
        初始化插件
        :param app:
        :return:
        """
        raise NotImplemented()

    @abstractmethod
    def destroy(self):
        """
        销毁插件
        :return:
        """
        raise NotImplemented()



none = 0
handle_coming = 1 << 0
handle_leaving = 1 << 2
handle_request = 1 << 3
before_invoke = 1 << 4
after_return = 1 << 5

all = handle_coming | handle_leaving | handle_request | before_invoke | after_return



import os
import tempfile
from contextlib import contextmanager
from types import FunctionType
from typing import List, Optional, Tuple

from .interfaces import IDbSessionProvider, ISessionProvider
from .session import HttpSession


class MemorySessionProvider(ISessionProvider):
    """
    基于内存的 session 管理
    注：此类型仅适用于单线程开发环境, 多线程时会出现无法预料的问题；请勿用于生产环境！
    """

    def __init__(self, expired: int = None, check_interval=0, on_expired: FunctionType = None):
        super().__init__(expired, check_interval, True, on_expired)
        self.sessions = {}

    def remove(self, session_id: str):
        if self.exists(session_id):
            del self.sessions[session_id]

    def clear(self):
        self.sessions.clear()

    def get_expired_session(self, time_before: float) -> List[str]:
        expired_sessions = []
        for session in self.sessions.values():
            if time_before > session.last_access_time:
                expired_sessions.append(session.id)
        return expired_sessions

    def get(self, session_id: str) -> Optional[HttpSession]:
        return self.sessions[session_id] if self.exists(session_id) else None

    def set(self, session: HttpSession):
        self.sessions[session.id] = session

    def exists(self, session_id: str):
        return session_id in self.sessions


class FileSessionProvider(ISessionProvider):
    def __init__(self, sessions_root: str = None, expired: int = None, check_interval=0, auto_clear=False,
                 on_expired: FunctionType = None):
        super().__init__(expired, check_interval, auto_clear, on_expired)

        if not sessions_root:
            sessions_root = tempfile.gettempdir()

        self.sessions_root = os.path.abspath(os.path.join(sessions_root, 'restfx_sessions'))
        self.logger.info('Session storage root: %s', self.sessions_root)

        if not os.path.exists(self.sessions_root):
            os.makedirs(self.sessions_root)
            # print('mkdir:' + self.sessions_root)
        elif auto_clear:
            self.clear()

    def _get_session_path(self, session_id: str) -> str:
        # session_id 中可能存在 / 符号
        return os.path.join(self.sessions_root, session_id.replace('/', '_'))

    def _load_session(self, session_id: str):
        # print('Load session: ' + session_id)
        session_file = self._get_session_path(session_id)
        if not os.path.isfile(session_file):
            # print('The session currently loading not exists:' + session_file)
            return None

        if os.path.getsize(session_file) == 0:
            # print('The session currently loading is empty:' + session_file)
            return None

        import pickle

        with open(session_file, mode='rb') as fp:
            try:
                session = pickle.load(fp)
                fp.close()
            except Exception as e:
                self.logger.error('Failed to load session file %s' % session_file, excpt=e)
                fp.close()
                # 无法解析 session 文件
                # 说明session文件已经损坏，将其删除
                os.remove(session_file)
                return None

            setattr(session, '_update_watcher', self.set)
            setattr(session, '_drop_watcher', self.remove)
            return session

    def remove(self, session_id: str):
        # print('Remove session:' + session_id)
        session_file = self._get_session_path(session_id)
        if not self.exists(session_file):
            # print("The session to remove is not exists:" + session_file)
            return
        os.remove(session_file)

    def get_expired_session(self, time_before: float) -> List[str]:
        entities = os.listdir(self.sessions_root)

        sessions = []

        for entity in entities:
            last_access_time = os.path.getatime(self._get_session_path(entity))
            if time_before > last_access_time:
                sessions.append(entity)

        return sessions

    def get(self, session_id: str) -> Optional[HttpSession]:
        return self._load_session(session_id)

    def set(self, session: HttpSession):
        import pickle
        import fcntl
        # print('Set session:' + session.id)
        fp = open(self._get_session_path(session.id), mode='wb')
        fcntl.flock(fp, fcntl.LOCK_EX)
        try:
            pickle.dump(session, fp, pickle.HIGHEST_PROTOCOL)
            fp.close()
        finally:
            fcntl.flock(fp, fcntl.LOCK_UN)

    def exists(self, session_id: str):
        return os.path.isfile(self._get_session_path(session_id))

    def clear(self):
        os.removedirs(self.sessions_root)


class MySQLSessionProvider(IDbSessionProvider):
    def __init__(self, db_options: dict, table_name="restfx_sessions", expired: int = None, check_interval=0,
                 auto_clear=False, on_expired: FunctionType = None):
        """

        :param db_options: 按 `pymysql.connect` 的参数传入: host, user, password, database, port, charset 等
        :param table_name:
        :param expired:
        :param check_interval:
        :param auto_clear:
        """
        self.table_name = table_name
        super().__init__({
            'connect_timeout': 5,
            **db_options
        }, expired, check_interval, auto_clear, on_expired)

    def execute(self, sql: str, *args, throw_except=False, ensure_table=True) -> Tuple[int, Tuple[dict]]:
        if ensure_table:
            self.ensure_table()

        import pymysql
        conn = pymysql.connect(**self.db_options)
        # print('[MySQLSessionProvider] ' + (sql % args))

        is_select = sql.startswith('SELECT')

        cursor = None
        rows = 0
        data = None
        try:
            import pymysql.cursors
            cursor = conn.cursor(pymysql.cursors.DictCursor)
            rows = cursor.execute(sql, args)
            if is_select:
                data = cursor.fetchall()
        except Exception as e:
            self.logger.error('Error occurred during executing SQL statement: %s, args=%s' % (sql, args), e)
            if throw_except:
                raise e
        else:
            if not is_select:
                conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            conn.close()

        # noinspection PyTypeChecker
        return rows, data

    def ensure_table(self):
        if self.is_db_available:
            return

        rows, _ = self.execute(
            """SELECT * FROM `information_schema`.`TABLES`
            WHERE `TABLE_SCHEMA`='{db_name}' AND `TABLE_NAME` ='{table_name}' LIMIT 1""".format(
                db_name=self.db_options['database'],
                table_name=self.table_name
            ), throw_except=True, ensure_table=False)
        if rows > 0:
            self.is_db_available = True
            if self.auto_clear:
                self.clear()
            return

        self.logger.info('Creating session table %r' % self.table_name)
        self.execute("""CREATE TABLE `{table_name}` (
        `id` VARCHAR(48) PRIMARY KEY NOT NULL,
        `creation_time` BIGINT NOT NULL,
        `last_access_time` BIGINT NOT NULL,
        `store` BLOB,
        INDEX `last_access`(`last_access_time` ASC)
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8""".format(table_name=self.table_name), throw_except=True,
                     ensure_table=False)

        self.is_db_available = True

    def get_expired_session(self, time_before: float) -> List[str]:
        rows, data = self.execute(
            'SELECT `id` FROM `{table_name}` WHERE `last_access_time` < %s'.format(table_name=self.table_name),
            int(time_before))
        if rows == 0:
            return []
        return [item['id'] for item in data]

    def get(self, session_id: str) -> Optional[HttpSession]:
        rows, data = self.execute(
            "SELECT * FROM `{table_name}` WHERE `id`=%s LIMIT 1".format(table_name=self.table_name),
            session_id)
        if rows == 0:
            return None
        return self.parse(data[0])

    def upsert(self, session_id: str, creation_time: float, last_access_time: float, store: bytes):
        self.execute("""INSERT INTO `{table_name}` VALUES(%s, %s, %s, %s) ON DUPLICATE KEY UPDATE
        last_access_time=%s, store=%s""".format(table_name=self.table_name),
                     session_id,
                     int(creation_time),
                     int(last_access_time),
                     store,
                     int(last_access_time),
                     store)

    def exists(self, session_id: str) -> bool:
        rows, _ = self.execute("""SELECT 1 FROM `{table_name}` WHERE `id`=%s limit 1""".format(
            table_name=self.table_name), session_id)
        return rows > 0

    def remove(self, session_id: str):
        self.execute("""DELETE FROM `{table_name}` WHERE `id`=%s""".format(table_name=self.table_name), session_id)

    def clear(self):
        self.execute('TRUNCATE TABLE `{table_name}`'.format(table_name=self.table_name))


class SqliteSessionProvider(IDbSessionProvider):
    def __init__(self, db_options: dict = None, table_name="restfx_sessions", expired: int = None, check_interval=0,
                 auto_clear=False, on_expired: FunctionType = None):
        """

        :param db_options: 按 `sqlite.connect` 的参数传入: database, timeout 等；``database`` 指定为 ``:memory:``(默认值) 表示使用内存数据库
        :param table_name:
        :param expired:
        :param check_interval:
        :param auto_clear:
        """
        self.table_name = table_name
        if not db_options:
            db_options = {}
        super().__init__({
            'timeout': 5,
            'database': ':memory:',
            **db_options,
            'check_same_thread': False
        }, expired, check_interval, auto_clear, on_expired)

        self.is_memory_mode = self.db_options['database'] == ':memory:'
        self.conn = None

    @classmethod
    def _row_parser(cls, cursor, row):
        item = {}
        for idx, col in enumerate(cursor.description):
            item[col[0]] = row[idx]
        return item

    def execute(self, sql: str, *args, throw_except=False, ensure_table=True) -> Tuple[int, Tuple[dict]]:
        if ensure_table:
            self.ensure_table()

        if not self.is_memory_mode or not self.conn:
            import sqlite3
            self.conn = sqlite3.connect(**self.db_options)
            self.conn.row_factory = self._row_parser

        conn = self.conn

        # print('[SqliteSessionProvider] %s\n%s' % (sql, args))

        is_select = sql.startswith('SELECT')

        cursor = None
        rows = 0
        data = None
        try:
            cursor = self.conn.execute(sql, args)
            if is_select:
                data = cursor.fetchall()
                rows = len(data)
            else:
                rows = cursor.rowcount
        except Exception as e:
            self.logger.error('Error occurred during executing SQL statement: %s, args=%s' % (sql, args), e)
            if throw_except:
                raise e
        finally:
            if cursor is not None:
                cursor.close()

            # 使用内存模式时，不能关闭连接
            # 否则内存的数据会在下次使用时丢失
            # 程序关闭时会自动处理这些问题
            if not self.is_memory_mode:
                conn.close()

        # noinspection PyTypeChecker
        return rows, data

    def ensure_table(self):
        if self.is_db_available:
            return

        rows, _ = self.execute(
            """SELECT * FROM `sqlite_master`
            WHERE `type`='table' AND `name` = '{table_name}' LIMIT 1""".format(
                table_name=self.table_name
            ), throw_except=True, ensure_table=False)
        if rows > 0:
            self.is_db_available = True
            if self.auto_clear:
                self.clear()
            return

        self.logger.info('Creating session table %r' % self.table_name)
        self.execute("""CREATE TABLE `{table_name}` (
        `id` VARCHAR(48) PRIMARY KEY NOT NULL,
        `creation_time` BIGINT NOT NULL,
        `last_access_time` BIGINT NOT NULL,
        `store` BLOB
        )""".format(table_name=self.table_name), throw_except=True,
                     ensure_table=False)
        # 创建索引
        self.execute("""CREATE INDEX `last_access` ON `{table_name}`(`last_access_time`)""".format(
            table_name=self.table_name
        ), throw_except=True, ensure_table=False)
        self.is_db_available = True

    def get_expired_session(self, time_before: float) -> List[str]:
        rows, data = self.execute(
            'SELECT `id` FROM `{table_name}` WHERE `last_access_time` < ?'.format(table_name=self.table_name),
            int(time_before))
        if rows == 0:
            return []
        return [item['id'] for item in data]

    def get(self, session_id: str) -> Optional[HttpSession]:
        rows, data = self.execute(
            "SELECT * FROM `{table_name}` WHERE `id`=? LIMIT 1".format(table_name=self.table_name),
            session_id)
        if rows == 0:
            return None
        return self.parse(data[0])

    def upsert(self, session_id: str, creation_time: float, last_access_time: float, store: bytes):
        self.execute("""REPLACE INTO `{table_name}`(`id`, `creation_time`, `last_access_time`, `store`)
         VALUES(?, ?, ?, ?)""".format(table_name=self.table_name),
                     session_id,
                     int(creation_time),
                     int(last_access_time),
                     store)

    def exists(self, session_id: str) -> bool:
        rows, _ = self.execute("""SELECT 1 FROM `{table_name}` WHERE `id`=? limit 1""".format(
            table_name=self.table_name), session_id)
        return rows > 0

    def remove(self, session_id: str):
        self.execute("""DELETE FROM `{table_name}` WHERE `id`=?""".format(table_name=self.table_name), session_id)

    def clear(self):
        self.execute('TRUNCATE TABLE `{table_name}`'.format(table_name=self.table_name))


class RedisSessionProvider(IDbSessionProvider):
    """
    提供基于 Redis 的 session 存储支持
    """

    def __init__(self, db_options: dict, expired: int = None, check_interval=0, auto_clear=False,
                 on_expired: FunctionType = None):
        """
        :param db_options: 按 `redis.connection.Connection` 的参数传入: host, password, port=6379, db=0 等
        :param expired:
        :param check_interval:
        :param auto_clear:
        """
        super().__init__({
            'socket_timeout': 5,
            **db_options
        }, expired, check_interval, auto_clear, on_expired)
        self.started = False

    @contextmanager
    def connect(self):
        import redis

        conn = redis.Redis(**self.db_options)
        if not self.started and self.auto_clear:
            self.started = True
            conn.flushdb()
        try:
            yield conn
        finally:
            conn.close()

    def get_expired_session(self, time_before: float) -> List[str]:
        return []

    def get(self, session_id: str) -> Optional[HttpSession]:
        with self.connect() as conn:
            session = conn.get(session_id)
            if session is None:
                return session

            import pickle
            session = pickle.loads(session)
            setattr(session, '_update_watcher', self.set)
            setattr(session, '_drop_watcher', self.remove)
            return session

    def upsert(self, session_id: str, creation_time: float, last_access_time: float, store: bytes):
        raise NotImplemented()

    def set(self, session: HttpSession):
        with self.connect() as conn:
            import pickle
            buf = pickle.dumps(session, pickle.HIGHEST_PROTOCOL)
            with conn.lock('set_session'):
                result = conn.set(session.id, buf, ex=self.expired)
            return result

    def exists(self, session_id: str) -> bool:
        with self.connect() as conn:
            return conn.exists(session_id) > 0

    def remove(self, session_id: str):
        with self.connect() as conn:
            return conn.delete(session_id)

    def clear(self):
        with self.connect() as conn:
            return conn.flushdb()



import logging
from typing import Tuple

from ..http import HttpRequest, HttpResponse
from . import MiddlewareBase


class StatusFilter:
    def __init__(self,
                 eq: int = None,
                 lt: int = None,
                 lte: int = None,
                 gt: int = None,
                 gte: int = None,
                 include: Tuple[int] = None,
                 exclude: Tuple[int] = None
                 ):
        self.eq = eq
        self.lt = lt
        self.lte = lte
        self.gt = gt
        self.gte = gte
        self.include = include
        self.exclude = exclude

    def set_eq(self, status: int):
        self.eq = status
        return self

    def set_lt(self, status: int):
        self.lt = status
        return self

    def set_lte(self, status: int):
        self.lte = status
        return self

    def set_gt(self, status: int):
        self.gt = status
        return self

    def set_gte(self, status: int):
        self.gte = status
        return self

    def set_include(self, status: Tuple[int]):
        self.include = status
        return self

    def set_exclude(self, status: Tuple[int]):
        self.exclude = status
        return self


class AccessMiddleware(MiddlewareBase):
    def __init__(self, logger_name: str = 'restfx', status_filter: StatusFilter = None):
        self.logger_name = logger_name
        self.logger = logging.getLogger(self.logger_name)
        self.status_filter = status_filter or StatusFilter()

    def on_leaving(self, request: HttpRequest, response: HttpResponse):
        status = response.status_code

        status_filter = self.status_filter

        if status_filter.lt is not None:
            if status >= status_filter.lt:
                return
        if status_filter.lte is not None:
            if status > status_filter.lte:
                return
        if status_filter.gt is not None:
            if status <= status_filter.gt:
                return
        if status_filter.gte is not None:
            if status < status_filter.gte:
                return
        if status_filter.include is not None:
            if status not in status_filter.include:
                return
        if status_filter.exclude is not None:
            if status in status_filter.exclude:
                return

        log_tpl = '[access] {remote} "{method} {path}" {status} {{{length}}}'.format(
            remote=request.environ.get('REMOTE_ADDR'),
            method=request.method,
            path=request.path,
            status=status,
            length=response.content_length or '-'
        )

        if status >= 500:
            self.logger.error(log_tpl)
        elif status >= 400:
            self.logger.warning(log_tpl)
        else:
            self.logger.debug(log_tpl)



import os

from ..config import AppConfig
from ..http import BadRequest, HttpResponse, JsonResponse, NotFound
from ..util import utils
from ..util.func_util import FunctionDescription


class ApiPage:
    def __init__(self, config: AppConfig):
        self.config = config
        # 开发模式的模块缓存，用于API列表
        self.routes_cache = {}
        # API列表页面缓存
        self.api_page_html_cache = ''

        self.prefix = '/%s%s' % (config.api_prefix, '/' if config.append_slash else '')

    def dispatch(self, request):
        if not self.config.api_page_enabled:
            from ..util import Logger
            Logger.current().info(
                'API page is disabled, '
                'use "App(..., api_page_enabled=True, ...)" to enable it.')
            return NotFound()

        prefix = self.prefix

        if request.path.lower() == prefix:
            return self.render_page(request)

        if not prefix.endswith('/'):
            prefix += '/'

        if request.path.lower() == prefix + 'export':
            return self.do_export(request)

        if request.path.lower() == prefix + 'api.json':
            return self.return_api_info(request)

        return NotFound()

    def render_page(self, request):
        if request.method != 'GET':
            return NotFound()
        if not self.api_page_html_cache or not self.config.api_page_cache:
            from .. import __meta__
            with open(os.path.join(os.path.dirname(__file__),
                                   '../internal_assets/templates/api_page.html'),
                      encoding='utf-8') as fp:
                lines = [
                    line.replace('{{version}}', __meta__.version)
                    for line in fp.readlines()
                ]
                self.api_page_html_cache = ''.join(lines)
                fp.close()
        return HttpResponse(self.api_page_html_cache, content_type='text/html')

    def return_api_info(self, request):
        if not self.routes_cache or not self.config.api_page_cache:
            self.routes_cache = self.config.docs

        from .. import __meta__
        return JsonResponse({
            'api_version': __meta__.api_version,
            'meta': {
                'name': __meta__.name,
                'version': __meta__.version,
                'url': __meta__.website
            },
            'name': self.config.api_page_name,
            'expanded': self.config.api_page_expanded,
            'routes': self.routes_cache,
            'enums': [
                utils.get_enum_items(item)
                for item in self.config.enum_types
            ],
            'custom_assets': self.config.api_page_assets
        }, encoder=FunctionDescription.JSONEncoder, ensure_ascii=request.GET.get('ascii') == 'true')

    def do_export(self, request):
        if request.method != 'POST':
            return NotFound()

        if 'data' not in request.POST:
            return BadRequest()

        content = request.POST['data']

        # noinspection PyBroadException
        try:
            from ..util import b64
            content = b64.dec_bytes(content)
        except Exception:
            # This exception can be ignored safety
            return BadRequest()

        import time

        from ..http import FileResponse
        return FileResponse(content,
                            content_type='application/markdown;charset=utf-8',
                            request=request,
                            attachment='%s[%s].md' % (
                                self.config.api_page_name,
                                time.strftime('%Y%m%I%H%M%S'))
                            )
